/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <linux/string.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/reiserfs_fs.h>
#include <linux/stat.h>
#include <asm/uaccess.h>

#else

#include "nokernel.h"

#endif



/*#ifdef __KERNEL__*/

extern struct key  MIN_KEY;

static ssize_t reiserfs_dir_read  (struct file * filp, char * buf, size_t count, loff_t * ppos)
{
  return -EISDIR;
}

static int reiserfs_readdir (struct file *, void *, filldir_t);

static struct file_operations reiserfs_dir_operations = {
	NULL,			/* lseek - default */
	reiserfs_dir_read,	/* read */
	NULL,			/* write */
	reiserfs_readdir,	/* readdir */
	NULL,			/* poll - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap */
	NULL,			/* no special open code */
	NULL,			/* flush */
	NULL,			/* no special release code */
	file_fsync,		/* default fsync */
	NULL, 			/* fasync */
	NULL,			/* check_media_change */
	NULL,			/* revalidate */
	NULL			/* lock */
};

/*
 * directories can handle most operations...
 */
struct inode_operations reiserfs_dir_inode_operations = {
	&reiserfs_dir_operations,	/* default directory file-ops */
	reiserfs_create,		/* create */
	reiserfs_lookup,		/* lookup */
	0/*reiserfs_link*/,			/* link */
	reiserfs_unlink,		/* unlink */
	0/*reiserfs_symlink*/,		/* symlink */
	reiserfs_mkdir,			/* mkdir */
	reiserfs_rmdir,			/* rmdir */
	0/*reiserfs_mknod*/,			/* mknod */
	0/*reiserfs_rename*/,		/* rename */
	NULL,				/* readlink */
	NULL,				/* follow_link */
	NULL,				/* readpage */
	NULL,				/* writepage */
	NULL,				/* bmap */
	NULL,				/* truncate */
	NULL				/* permission */
};



/* directory can have few entries with the same hash value. But only on of them or none is visible. */
static int reiserfs_readdir (struct file * filp, void * dirent, filldir_t filldir)
{
  struct inode *inode = filp->f_dentry->d_inode;
  struct key entry_key;	/* key of current entry itemposition in the directory (key of directory entry) */
  struct buffer_head * bh;
  int item_num;
  int repeat;
  struct key * rkey;
  struct item_head * ih;

  struct path path_to_entry;

  init_path (&path_to_entry);

  if (!inode || !inode->i_sb || !S_ISDIR (inode->i_mode))
    return -EBADF;

  if (filp->f_pos == 0) {
    if (filldir (dirent, ".", strlen ("."), filp->f_pos, inode->i_ino) < 0)
      return 0;
    filp->f_pos = 1/*DOT_DOT_OFFSET*/;
  }

  if (filp->f_pos == 1/*DOT_DOT_OFFSET*/) {
    if (filldir (dirent, "..", strlen (".."), filp->f_pos, INODE_PARENT_DIR_ID (inode)/* objectid of the ".." */) < 0)
      return 0;
    filp->f_pos = 1/*DOT_DOT_OFFSET*/ + 1;
  }

  /* form key for search the next directory entry using f_pos field of file structure  */
  entry_key.group_locale.major = inode->i_ino;
  entry_key.group_locale.minor = 0;
  entry_key.name_hash = filp->f_pos;
  entry_key.file_id = 0;
  entry_key.k_offset = 0;

  while (1) {
    /* search the directory item, containing entry with specified key */
    search_by_key (inode->i_sb, &entry_key, &path_to_entry, &repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS);

    bh = PATH_PLAST_BUFFER (&path_to_entry);
    item_num = PATH_LAST_POSITION (&path_to_entry);
    ih = B_N_PITEM_HEAD (bh, item_num);
		
    /* scan all items going to the right */
    while (item_num < B_NR_ITEMS (bh) &&	/* we are inside of block */
	   entry_key.group_locale.major == ih->ih_key.group_locale.major /* we are inside of directory being scanned */) {
      if (is_dir_entry (ih) && !is_entry_hidden (ih)) {
	if (filldir (dirent, B_I_NAME (bh, ih), B_I_NAMELEN (bh, ih), ih->ih_key.k_offset,
		     B_I_OBJECT_LOCALE_FROM_ENTRY (bh, ih)->file_id) < 0) {
	  pathrelse (&path_to_entry);
	  return 0;
	}
	/* set position in directory to the next entry */
	filp->f_pos = ih->ih_key.name_hash + 1;
      } else
	/* item is not entry or hidden entry */
	if (filp->f_pos < ih->ih_key.name_hash)
	  filp->f_pos = ih->ih_key.name_hash;
      item_num ++;
      ih ++;
    }

    if (item_num == B_NR_ITEMS (bh)) {
      /* we have reached end of the node. Get right delimiting key to check whether it is end of the
         directory */
      rkey = get_rkey (&path_to_entry, inode->i_sb);
      if (! comp_keys (rkey, &MIN_KEY)) {
#ifdef REISERFS_CHECK
	reiserfs_warning ("vs-9000: reiserfs_readdir: could not calculate rigth delimiting key (path has been changed)\n");
#endif
	/* set pos_key to key, that is the smallest and greater that key of the last entry in the
	   item */
	entry_key.name_hash = filp->f_pos;
	entry_key.file_id = 0;
	entry_key.k_offset = 0;
	continue;
      }
      if (rkey->group_locale.major != entry_key.group_locale.major) {
	/* end of the directory */
	pathrelse (&path_to_entry);
	return 0;
      }
      /* directory continues in the right neighboring block */
      copy_key (&entry_key, rkey);
      continue;
    }
    /* directory item is not a last item of the node. End of the directory is achieved */
    pathrelse (&path_to_entry);
    return 0;
  } /* while */

  return 0;
}

/*#endif  __KERNEL__ */




















