/*
 * print tree file system after umounting
 */
#include "../include/nokernel.h"
#include <unistd.h>
#include <stdlib.h>




/*
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <asm/bitops.h>
#include <linux/fs.h>

#include "../include/io.h"
#include "../include/sb.h"
*/
/*#include "../kernel-include/reiserfs_fs_sb.h"
#include "../kernel-include/reiserfs_fs.h"*/

void check_and_free_mem (void);
void die (char * fmt, ...);
int is_mounted (char * device_name);



#define usage() die ("Usage: %s [-b block-to-print] [-idcom] device", argv[0]);


struct reiserfs_fsstat {
  int nr_internals;
  int nr_leaves;
  int nr_directories;
  int nr_files;
  int nr_unformatted;
} g_stat_info;
int g_comp_number = 0;

/*
 *  options
 */
int opt_print_regular_file_content = 0;/* -c */
int opt_print_directory_contents = 0;	/* -d */
int opt_print_leaf_items = 0;		/* -i */
int opt_print_objectid_map = 0;	/* -o */
int opt_print_block_map = 0;		/* -m */
/* when you want print one block specify -b # */
int opt_block_to_print = -1;


struct super_block g_sb;

/*struct buffer_head * g_sbh;
struct reiserfs_super_block * g_rs;*/

/*
void read_super_block (int dev)
{

  g_sbh = bread (dev, REISERFS_FIRST_BLOCK, 1024, 0);
  g_rs = (struct reiserfs_super_block *)g_sbh->b_data;
  if (strncmp (g_rs->s_magic, REISERFS_SUPER_MAGIC, strlen (REISERFS_SUPER_MAGIC)))
    die ("read_super_block: can't find a reiser filesystem on dev %x\n", dev);
  g_sb.u.reiserfs_sb.s_rs = g_rs;
}
*/

int g_max_key_amount = 0;





struct key g_right_delim_key = {0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff};


int print_mode (void)
{
  int mode = 0;

  if (opt_print_leaf_items)
    mode |= PRINT_LEAF_ITEMS;
  if (opt_print_directory_contents)
    mode |= (PRINT_LEAF_ITEMS | PRINT_DIRECTORY_ITEMS);
  if (opt_print_regular_file_content)
    mode |= (PRINT_LEAF_ITEMS | PRINT_DIRECT_ITEMS);
  return mode;
}


void print_disk_tree (int block_nr)
{
  struct buffer_head * bh;

  bh = bread (g_sb.s_dev, block_nr, g_sb.s_blocksize);
  if (B_IS_KEYS_LEVEL (bh)) {
    int i;
    struct disk_child * dc;

    g_stat_info.nr_internals ++;
    print_block (bh, print_mode (), -1, -1);
      
    dc = B_N_CHILD (bh, 0);
    for (i = 0; i <= B_NR_ITEMS (bh); i ++, dc ++)
      print_disk_tree (dc->dc_block_number);

  } else if (B_IS_ITEMS_LEVEL (bh)) {
    g_stat_info.nr_leaves ++;

    if (B_PRIGHT_DELIM_KEY (bh)->file_id == 0xffffffff) {
      g_max_key_amount ++;
    }

    /*&&&&&&&&&&&&&&&*/
    if (g_right_delim_key.file_id != 0xffffffff) {
      if (comp_keys (&g_right_delim_key, B_N_PKEY (bh, 0))) {
	reiserfs_panic (0, "g_right_delim_key in previous leaf %k, 0th key in S %k",
			&g_right_delim_key, B_N_PKEY (bh, 0));
      }
      g_comp_number ++;
    }
    memcpy (&g_right_delim_key, B_PRIGHT_DELIM_KEY (bh), KEY_SIZE);
    /*&&&&&&&&&&&&&&&*/

    print_block (bh, print_mode (), -1, -1);
  } else {
    print_block (bh, print_mode (), -1, -1);
    die ("print_disk_tree: bad block type");
  }
  brelse (bh);
}


/* hlam */
struct buffer_head * g_left = 0;
struct buffer_head * g_right = 0;
struct key * g_dkey = 0;
int g_compares = 0;


void compare_keys ()
{
  g_compares ++;
  if (comp_keys (g_dkey, B_N_PKEY (g_right, 0)))
    die ("bad keys");
  if (comp_keys (g_dkey, B_PRIGHT_DELIM_KEY (g_left)))
    die ("bad keys");
}


struct buffer_head * find_buffer (int dev, int block, int size);
void scan_tree (int block)
{
  struct buffer_head * bh;

  bh = find_buffer (0, block, g_sb.s_blocksize);
  if (bh == 0) {
    g_left = 0; g_right = 0; g_dkey = 0;
    return;
  }
  bh->b_count ++;

  if (B_IS_KEYS_LEVEL (bh)) {
    int i;
    struct disk_child * dc;

    dc = B_N_CHILD (bh, 0);
    for (i = 0; i <= B_NR_ITEMS (bh); i ++, dc ++) {
      scan_tree (dc->dc_block_number);
      g_dkey = B_N_PDELIM_KEY (bh, i);
    }
  } else if (B_IS_ITEMS_LEVEL (bh)) {
    g_right = bh;
    if (g_left != 0 && g_dkey != 0)
      compare_keys ();
    g_left = g_right;
  } else {
    die ("bad block type");
  }
  brelse (bh);
}
/* end of hlam */


void print_one_block (int block)
{
  struct buffer_head * bh;
    
  bh = bread (g_sb.s_dev, block, g_sb.s_blocksize);
  print_block (bh, PRINT_LEAF_ITEMS | PRINT_DIRECTORY_ITEMS | (opt_print_regular_file_content ? PRINT_DIRECT_ITEMS : 0), -1, -1);
  brelse (bh);
  check_and_free_mem ();
  return;
}


static char * parse_options (int argc, char * argv [])
{
  int c;
  char * tmp;
  
  while ((c = getopt (argc, argv, "b:icdom")) != EOF) {
    switch (c) {
    case 'b':	/* print a single node */
      opt_block_to_print = strtol (optarg, &tmp, 0);
      if (*tmp)
	die ("parse_options: bad block size");
      break;

    case 'i':	/* print items of a leaf */
      opt_print_leaf_items = 1; break;

    case 'd':	/* print directories */
      opt_print_directory_contents = 1; break;

    case 'c':	/* print contents of a regular file */
      opt_print_regular_file_content = 1; break;

    case 'o':	/* print a objectid map */
      opt_print_objectid_map = 1; break;

    case 'm':	/* print a block map */
      opt_print_block_map = 1;  break;
    }
  }
  if (optind != argc - 1)
    /* only one non-option argument is permitted */
    usage ();
  
  return argv[optind];
}


int main (int argc, char * argv[])
{
  char * file_name;
  int dev;

#if 0
  if (0) {
    /* ???? */
    getblk (0,0,0);
    iput (0);
  }
#endif

  printf ("\n<-----------PRINT_DISK_LAYOUT, version 0.91a, 1999----------->\n");
 
  file_name = parse_options (argc, argv);

  init_buffer_cache ();

  dev = open (file_name, O_RDONLY);
  if (dev == -1)
    die ("print_disk_layout: Can not open device %s: %s\n", file_name, strerror (errno));
  read_super_block (dev, &g_sb);
  
  if (opt_block_to_print != -1) {
    print_one_block (opt_block_to_print);
    return 0;
  }

  print_super_block (&g_sb, "SUPER BLOCK");

  if (opt_print_objectid_map)
    print_objectid_map (&g_sb);

  if (opt_print_block_map) {
    int j;
    struct buffer_head ** bhs = 0;
    char ** bmap = 0;

    read_true_bitmap (&g_sb, &bhs, &bmap);
    print_bmap (SB_BMAP_NR (&g_sb), bmap, g_sb.s_blocksize, "TRUE BITMAP");
    
    for (j = 0; j < SB_BMAP_NR (&g_sb); j ++)
      brelse (bhs[j]);

    read_cautious_bitmap (&g_sb, &bhs, &bmap);
    print_bmap (SB_BMAP_NR (&g_sb), bmap, g_sb.s_blocksize, "CAUTIOUS BITMAP");
    for (j = 0; j < SB_BMAP_NR (&g_sb); j ++)
      brelse (bhs[j]);
    freemem (bhs, sizeof (struct buffer_head *) * SB_BMAP_NR (&g_sb));
    freemem (bmap, sizeof (char *) * SB_BMAP_NR (&g_sb));
  }

  

  /* print file system tree */
  print_disk_tree (SB_ROOT_BLOCK (&g_sb));

  /* print the statistic */
  printf ("File system on \"%s\" contains: %d files, %d directories\n",
	  file_name, g_stat_info.nr_files, g_stat_info.nr_directories);
  printf ("File systen uses %d internal + %d leaves + %d unformatted nodes = %d blocks\n", 
	  g_stat_info.nr_internals, g_stat_info.nr_leaves, g_stat_info.nr_unformatted, 
	  g_stat_info.nr_internals + g_stat_info.nr_leaves + g_stat_info.nr_unformatted);

  brelse (SB_BUFFER_WITH_SB (&g_sb));
  check_and_free_mem ();

  return 0;
}
/* end of main */



